class NonlocalBoxExceptions(Exception):
    ...


class ServiceError(NonlocalBoxExceptions):
    ...


class StatusError(NonlocalBoxExceptions):
    ...


class UninitializedBoxError(NonlocalBoxExceptions):
    ...



from requests import get

from .exceptions import *
from .settings import API_URL

__all__ = ["NonlocalBox"]


class NonlocalBox:
    """
    An API for simulating nonlocal no-signalling correlations.

    Attributes
    ----------
    api_key : str
        The API key to use the service.
    box_id : int
        The ID of the box use for the simulation(s).
    box_type_id : int
        The ID of the type of the box.
    role : str
        The role in the simulation. Can be Alice of Bob

    Raises
    ------
    TypeError
        If *api_key* is not a string.
    """
    def __init__(self, api_key):
        self.api_key = api_key
        self.box_id = None
        self.box_type_id = None
        self.role = None
        if not isinstance(self.api_key, str):
            raise TypeError("api_key must be a string")

    def invite(self, partner, box_type_id, box_name):
        """
        Invite *partner* for the game, set `self.box_id` to the returned box_id
        and set the type of *box_id* to `self.box_type_id`.

        Parameters
        ----------
        partner : str
            The username of the partner to play with
        box_type_id : int
            The ID of the box to use in the game
        box_name : str
            A user defined name of the box

        Returns
        -------
        box_id : int
            The ID of the newly created box named *box_name*.

        Raises
        ------
        `~nonlocalbox.exception.StatusError`
            If a nonzero status value returned by the server.
        """
        r = get(API_URL + "/invitePartner",
                params={
                    "apiKey": self.api_key,
                    "inviteUserName": partner,
                    "boxTypeID": box_type_id,
                    "boxName": box_name,
                })
        _check_status(r)
        self.box_id = r.json()["boxID"]
        self.box_type_id = box_type_id
        return self.box_id

    def initialize(self, box_id, i_want_to_be_bob=False):
        """
        Set *box_id* without verification and set `self.role` to Alice
        if it is in the list of boxes; set it to Bob otherwise.

        Parameters
        ----------
        box_id : int
            The ID of the box used in the game.
        i_want_to_be_bob : bool, optional
            If aliceUser and bobUser is the same in the box with box ID *box_id*
            this boolean will tell the server that the user wants to be in the role Bob.
        """
        self.box_id = box_id
        box = next((i for i in self.list_boxes() if i["id"] == str(box_id)), None)
        if box is None:
            self.role = "Bob"
        elif box["aliceUser"] == box["bobUser"]:
            self.role = "Bob" if i_want_to_be_bob else "Alice"
        else:
            self.role = "Alice"

    def use(self, input, transaction_id):
        """Use box with *input* and *transaction_id*.

        Parameters
        ----------
        input : int
            The input of the box.
        transaction_id : str
            The transaction ID of the game.

        Returns
        -------
        res : int
            The output of the box.

        Raises
        ------
        `~nonlocalbox.exceptions.StatusError`
            If a nonzero status value returned by the server.
        `~nonlocalbox.exceptions.UnknownBoxTypeError`
            If box_id does not exist in the database.
        """
        if self.box_id is None:
            raise UninitializedBoxError(f"No box initialized.")
        r = get(API_URL + "/useBox",
                params={
                    "apiKey": self.api_key,
                    "boxID": self.box_id,
                    "transactionID": transaction_id,
                    "x" if self.role == "Alice" else "y": input,
                })
        _check_status(r)
        return r.json()["a"] if self.role == "Alice" else r.json()["b"]

    def list_box_types(self):
        """
        List all available box types.

        Returns
        -------
        r : list
            A list of box types

        Raises
        ------
        `~nonlocalbox.exceptions.StatusError`
            If the server returns a nonzero status.
        """
        r = get(API_URL + "/listBoxTypes",
                params={"apiKey": self.api_key})
        _check_status(r)
        return r.json()["boxTypes"]

    def list_boxes(self):
        """
        List all boxes where the user has role 'Alice'.

        Returns
        -------
        boxes : list
            A list of boxes where the user has role 'Alice'.

        Raises
        ------
        `~nonlocalbox.exceptions.StatusError`
            If a nonzero status value returned by the server.
        """
        r = get(API_URL + "/listBoxes",
                params={"apiKey": self.api_key})
        _check_status(r)
        return r.json()["boxes"]

    def box_type_info(self):
        """
        Return some information on box type stored in ``self``.

        Returns
        -------
        info : dict
            All information available on box type.

        Raises
        ------
        `~nonlocalbox.exceptions.StatusError`
            If a nonzero status value returned by the server.
        """
        r = get(API_URL + "/boxTypeInfo",
                params={
                    "apiKey": self.api_key,
                    "boxTypeID": self.box_type_id,
                })
        _check_status(r)
        return r.json()["boxTypes"]


def _check_status(resp):
    if resp.status_code != 200:
        raise ServiceError(f"Received an invalid status code from the server: {resp.status_code}.")
    if resp.json()["status"] != 0:
        raise StatusError(f"Nonzero status returned: {resp.json()['status']}")


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonlocalbox']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.27.1']

setup_kwargs = {
    'name': 'nonlocalbox',
    'version': '0.1',
    'description': 'A friendly API for simulating nonlocal no-signalling correlations',
    'long_description': 'NonlocalBox\n===========\n\n**A friendly API for simulating nonlocal no-signalling correlations.**\n\nInstallation\n------------\n\nThe easiest way to install ``NonlocalBox`` is to use the ``pip`` command:\n\n.. code:: sh\n\n    python -m pip install nonlocalbox\n\nYou may need to replace ``python`` with the correct Python interpreter, e.g., ``python3``.\n\nUsage\n-----\nThe following example illustrates a simple game between Alice and Bob. First, we\nneed to create two instances of ``NonlocalBox``, one for Alice and one for Bob:\n\n.. code:: python\n\n    from os import environ\n    from nonlocalbox import NonlocalBox\n\n    alice_game = NonlocalBox(environ["ALICE_API_KEY"])\n    bob_game = NonlocalBox(environ["BOB_API_KEY"])\n\nIn the current state, neither of them are in any role. Suppose that Alice invites\nBob for a simulation, whose username is known by Alice (which is \'bob\' in this case).\nAlice wants to use Popescu-Rohrlich Box of box ID 1 and names it \'hellothere\':\n\n.. code:: python\n\n    alice_game.invite("bob", 1, \'hellothere\')\n    print(alice_game.box_id)  # this is arbitrary\n    4\n\nIn the server side, Bob is automatically added to this box. They both should\ninitialize the newly created box with ID 4. This will set the role \'Alice\'\nto Alice and \'Bob\' to Bob (since there won\'t be any box in Bob\'s list with ID 4):\n\n.. code:: python\n\n    alice_game.initialize(4)\n    bob_game.initialize(4)\n\nThey can use the nonlocal boxes to run a simulation.\n\nSuppose Alice sends `x = 0` are her input to the box with transaction ID ``20220311001`` and\nBob sends `y = 0` with the same transaction ID. Note that for `x = y = 0` the results should\nbe correlated:\n\n.. code:: python\n\n    print(alice_game.use(0, "20220311001"))\n    0\n    print(bob_game.use(0, "20220311001"))\n    0\n\nNow suppose Bob will be the first to send `y = 1` with an incremented transaction ID, and\nAlice also sends `x = 1`. The results should be anticorrelated:\n\n.. code:: python\n\n    print(bob_game.use(1, "20220311002"))\n    1\n    print(alice_game.use(1, "20220311002"))\n    0\n\n',
    'author': 'Ottó Hanyecz',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ohanyecz/nonlocalbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
